#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import unittest

from fabric.actor.core.kernel.client_reservation_factory import ClientReservationFactory
from fabric.actor.core.time.actor_clock import ActorClock
from fabric.actor.core.time.calendar.client_calendar import ClientCalendar
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.reservation_set import ReservationSet


class ClientCalendarTest(unittest.TestCase):
    Offset = 1000
    Length = 10

    def _get_calendar(self):
        clock = ActorClock(beginning_of_time=self.Offset, cycle_millis=self.Length)
        return ClientCalendar(clock=clock)

    def _make_reservation(self, *, id: str):
        return ClientReservationFactory.create(rid=ID(uid=id))

    def test_create(self):
        cal = self._get_calendar()
        self.assertIsNotNone(cal.clock)
        self.assertIsNotNone(cal.demand)
        self.assertIsNotNone(cal.holdings)
        self.assertIsNotNone(cal.pending)
        self.assertIsNotNone(cal.renewing)
        self.assertIsNotNone(cal.get_demand())
        self.assertIsNotNone(cal.get_holdings())
        self.assertIsNotNone(cal.get_pending())
        self.assertIsNotNone(cal.get_renewing(cycle=1000))

    def check_set(self, *, rset: ReservationSet, check: ReservationSet):
        self.assertIsNotNone(check)
        self.assertEqual(rset.size(), check.size())
        for r in rset.values():
            self.assertTrue(check.contains(reservation=r))

    def test_demand(self):
        cal = self._get_calendar()
        rset = ReservationSet()

        for i in range(5):
            r = self._make_reservation(id=str(i))
            # add to the list
            rset.add(reservation=r)
            cal.add_demand(reservation=r)
            # get the list and check it
            temp = cal.get_demand()
            self.check_set(rset=rset, check=temp)
            # remove from the returned set
            temp.remove(reservation=r)
            # make sure this did not affect the parent data structure
            temp = cal.get_demand()
            self.check_set(rset=rset, check=temp)

        for i in range(5):
            r = self._make_reservation(id=str(i))
            # add to the list
            rset.remove(reservation=r)
            cal.remove_demand(reservation=r)

            # get the list and check it
            temp = cal.get_demand()

            self.check_set(rset=rset, check=temp)

    def test_pending(self):
        cal = self._get_calendar()
        rset = ReservationSet()

        for i in range(5):
            r = self._make_reservation(id=str(i))
            rset.add(reservation=r)
            cal.add_pending(reservation=r)
            temp = cal.get_pending()
            self.check_set(rset=rset, check=temp)
            temp.remove(reservation=r)
            temp = cal.get_pending()
            self.check_set(rset=rset, check=temp)

        for i in range(5):
            r = self._make_reservation(id=str(i))
            rset.remove(reservation=r)
            cal.remove_pending(reservation=r)
            temp = cal.get_pending()
            self.check_set(rset=rset, check=temp)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING


from fabric.actor.core.apis.i_callback_proxy import ICallbackProxy
from fabric.actor.core.apis.i_delegation import IDelegation

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation
    from fabric.actor.core.apis.i_rpc_request_state import IRPCRequestState
    from fabric.actor.core.util.update_data import UpdateData
    from fabric.actor.security.auth_token import AuthToken


class IClientCallbackProxy(ICallbackProxy):
    @abstractmethod
    def prepare_update_ticket(self, *, reservation: IBrokerReservation, update_data: UpdateData,
                              callback: ICallbackProxy, caller: AuthToken) -> IRPCRequestState:
        """
        Prepare an update ticket
        @params reservation: reservation
        @params update_data: update data
        @params callback: callback
        @params caller: caller
        """

    @abstractmethod
    def prepare_update_delegation(self, *, delegation: IDelegation, update_data: UpdateData,
                                  callback: ICallbackProxy, caller: AuthToken) -> IRPCRequestState:
        """
        Prepare an update delegation
        @params delegation: delegation
        @params update_data: update data
        @params callback: callback
        @params caller: caller
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_actor import IActor
    from fabric.actor.core.apis.i_authority_reservation import IAuthorityReservation
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.actor.core.apis.i_reservation import IReservation
    from fabric.actor.core.kernel.resource_set import ResourceSet
    from fabric.actor.core.plugins.config.config_token import ConfigToken
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.util.resource_type import ResourceType


class IResourceControl:
    """
    Interface for authority policy resource control implementations. An authority
    policy organizes the authorities inventory into a number of resource pools.
    Each resource pool is associated with a resource control object responsible
    for servicing requests for resources for the resource pool.
    """

    @abstractmethod
    def donate_reservation(self, *, reservation: IClientReservation):
        """
        Informs the control about a source ticket. Depending on the control
        implementation, the control may choose to treat each source ticket as
        inventory.
        @params reservation reservation
        @raises Exception in case of error
        """

    @abstractmethod
    def donate(self, *, resource_set: ResourceSet):
        """
        Informs the control about physical inventory.
        @params resource_set set of inventory resources (inventory units)
        @raises Exception in case of error
        """

    @abstractmethod
    def eject(self, *, resource_set: ResourceSet):
        """
        Forcefully removes the specified resources from the control's inventory.
        @params resource_set set of resources to eject from the inventory (inventory
                   units)
        @raises Exception in case of error
        """

    @abstractmethod
    def release(self, *, resource_set: ResourceSet):
        """
        Releases previously allocated resources. Note that some of those
        resources may represent failed resource units. Failed units should not be
        released. Once the failure is corrected (by an entity external to the
        control), the control will be notified through freed
        @params resource_set set of released resources (allocated units)
        @raises Exception in case of error
        """

    @abstractmethod
    def available(self, *, resource_set: ResourceSet):
        """
        Notifies the control that inventory resources are now available for use.
        @params resource_set inventory resources that have become available (inventory units)
        @raises Exception in case of error
        """

    @abstractmethod
    def unavailable(self, *, resource_set: ResourceSet) -> int:
        """
        Indicates that some inventory resources should be marked as unavailable.
        @params resource_set resources (inventory units)
        @returns -1 if at least one resource unit is currently in use; 0 otherwise
        @raises Exception in case of error
        """

    @abstractmethod
    def freed(self, *, resource_set: ResourceSet):
        """
        Indicates that previously committed resources have been freed. Most
        likely these resources represent failed allocations.
        @params resource_set set of freed resources (allocated units)
        @raises Exception in case of error
        """

    @abstractmethod
    def failed(self, *, resource_set: ResourceSet):
        """
        Notifies the control that some inventory resources have failed and cannot
        be used to satisfy client requests.
        @params resource_set set of failed resources (inventory units)
        @raises Exception in case of error
        """

    @abstractmethod
    def recovered(self, *, resource_set: ResourceSet):
        """
        Notifies the policy that previously failed resources have been recovered
        and can be safely used to satisfy client requests.
        @params resource_set set of recovered resources
        @raises Exception in case of error
        """

    @abstractmethod
    def assign(self, *, reservation: IAuthorityReservation) -> ResourceSet:
        """
        Assigns resources to the reservation.
        @params reservation reservation
        @returns resources assigned to the reservation (allocated units)
        @raises Exception in case of error
        """

    @abstractmethod
    def correct_deficit(self, *, reservation: IAuthorityReservation) -> ResourceSet:
        """
        Informs the control that a reservation with previously allocated
        resources has a deficit. The deficit may be positive or negative. The
        control must decide how to handle the deficit. For example, the control
        may try allocating new resources.

        This method may be invoked multiple times. The core will invoke either
        until the deficit is corrected or the control indicates that the
        reservation must be sent back to the client even though it has a deficit.
        @params reservation reservation with a deficit
        @returns resources allocated/removed so that the deficit is corrected (allocated units)
        @raises Exception in case of error
        """

    @abstractmethod
    def close(self, *, reservation: IReservation):
        """
        Notifies the control that a reservation is about to be closed.
        This method will be invoked for every reservation that is about to be
        closed, even if the close was triggered by the control itself. The
        policy should update its internal state/cancel pending operations
        associated with the reservation. This method is invoked with the kernel
        lock on.
        @params reservation: reservation to close
        """

    @abstractmethod
    def recovery_starting(self):
        """
        Notifies the control that recovery is starting.
        """

    @abstractmethod
    def revisit(self, *, reservation: IReservation):
        """
        Recovers state for a reservation.
        @params reservation reservation
        @raises Exception in case of error
        """

    @abstractmethod
    def recovery_ended(self):
        """
        Notifies the control that recovery has ended.
        """

    @abstractmethod
    def configuration_complete(self, *, action: str, token: ConfigToken, out_properties: dict):
        """
        Notifies the control that a configuration action for the object
        represented by the token parameter has completed.

        @params action configuration action. See Config.Target*
        @params token object or a token for the object whose configuration action has completed
        @params out_properties output properties produced by the configuration action
        """

    @abstractmethod
    def get_guid(self) -> ID:
        """
        Returns the control guid. Every control must have a unique guid.
        @return resource control guid
        """

    @abstractmethod
    def get_types(self) -> set:
        """
        Returns the resource types this resource control handles. A resource
        control can be responsible for more than one resource type, although
        usually the mapping will be one-to-one.
        @returns list of resource types
        """

    @abstractmethod
    def add_type(self, *, rtype: ResourceType):
        """
        Instructs the control the handle the specified type.
        @param rtype type
        """

    @abstractmethod
    def remove_type(self, *, rtype: ResourceType):
        """
        Instructs the control it no longer handle the specified type.
        @param rtype type
        """

    @abstractmethod
    def register_type(self, *, rtype: ResourceType):
        """
        Registers the specified resource type with the resource control.
        @param rtype type
        """

    @abstractmethod
    def set_actor(self, *, actor: IActor):
        """
        Sets the actor the control is associated with.
        @params actor actor
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading

from fabric.actor.core.apis.i_event import IEvent
from fabric.actor.core.apis.i_event_handler import IEventHandler
from fabric.actor.core.apis.i_reservation_tracker import IReservationTracker
from fabric.actor.core.kernel.reservation_purged_event import ReservationPurgedEvent
from fabric.actor.core.kernel.reservation_state_transition_event import ReservationStateTransitionEvent
from fabric.actor.core.kernel.reservation_states import ReservationPendingStates, ReservationStates, JoinState
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.reservation_state import ReservationState


class ReservationTracker(IReservationTracker, IEventHandler):
    class ReservationTrackerState:
        def __init__(self):
            self.state = None
            self.purged = False

        def is_ticketing(self):
            """
            Return true if in ticketing state; false otherwise
            """
            return self.state.get_pending() == ReservationPendingStates.Ticketing or \
                   self.state.get_pending_state() == ReservationPendingStates.ExtendingTicket

        def is_ticketed(self):
            """
            Return true if in Ticketed state; false otherwise
            """
            return self.state.get_state() == ReservationStates.Ticketed or \
                   self.state.get_state() == ReservationStates.ActiveTicketed

        def is_extending_ticketing(self):
            """
            Return true if in Extending Ticketing state; false otherwise
            """
            return self.state.get_pending() == ReservationPendingStates.ExtendingTicket

        def is_active(self):
            """
            Return true if in Active state; false otherwise
            """
            if self.state.get_joining() == JoinState.None_:
                return self.state.get_state() == ReservationStates.Active and \
                       self.state.get_pending() == ReservationPendingStates.None_
            else:
                return (self.state.get_state() == ReservationStates.Active or
                        self.state.get_state() == ReservationStates.ActiveTicketed) and \
                       self.state.get_joining() == JoinState.NoJoin

        def is_closed(self):
            """
            Return true if in Closed state; false otherwise
            """
            return self.state.get_state() == ReservationStates.Closed

        def is_failed(self):
            """
            Return true if in Failed state; false otherwise
            """
            return self.state.get_state() == ReservationStates.Failed

        def is_terminal(self):
            """
            Return true if in Closed, failed or purged state; false otherwise
            """
            return self.is_closed() or self.is_failed() or self.purged

    def __init__(self):
        self.state = {}
        self.lock = threading.Lock()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['lock']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.lock = threading.Lock()

    def handle_state_transition(self, *, e: ReservationStateTransitionEvent):
        """
        Handle State transition event
        @param e state transition event
        """

        ts = None
        try:
            self.lock.acquire()
            if e.get_reservation_id() in self.state:
                ts = self.state.get(e.get_reservation_id(), None)
            if ts is None:
                ts = ReservationTracker.ReservationTrackerState()
                ts.state = e.get_state()
                self.state[e.get_reservation_id()] = ts
        finally:
            self.lock.release()

    def handle_reservation_purged(self, *, e: ReservationPurgedEvent):
        """
        Handle Purge event
        @param e purge event
        """
        ts = None
        try:
            self.lock.acquire()
            if e.get_reservation_id() in self.state:
                ts = self.state.pop(e.get_reservation_id())
                ts.purged = True
        finally:
            self.lock.release()

    def handle(self, *, event: IEvent):
        """
        Handle an incoming event
        @param event incoming event
        """
        if isinstance(event, ReservationStateTransitionEvent):
            self.handle_state_transition(e=event)
        elif isinstance(event, ReservationPurgedEvent):
            self.handle_reservation_purged(e=event)

    def get_state(self, *, rid: ID) -> ReservationState:
        """
        Return state for a reservation
        @param rid reservation id
        """
        ret_val = None
        try:
            self.lock.acquire()
            if rid in self.state:
                ret_val = self.state[rid]
        finally:
            self.lock.release()
        return ret_val



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import EventException
from fabric.actor.core.container.event_subscription import EventSubscription
from fabric.actor.core.container.synchronous_event_subscription import SynchronousEventSubscription

if TYPE_CHECKING:
    from fabric.actor.core.container.aevent_subscription import AEventSubscription
    from fabric.actor.security.auth_token import AuthToken
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.apis.i_event_handler import IEventHandler
    from fabric.actor.core.apis.i_event import IEvent


class EventManager:
    def __init__(self):
        self.subscriptions = {}

    def create_subscription(self, *, token: AuthToken, filters: list, handler: IEventHandler) -> ID:
        if token is None:
            raise EventException(Constants.invalid_argument)

        subscription = None
        if handler is None:
            subscription = EventSubscription(token=token, filters=filters)
        else:
            subscription = SynchronousEventSubscription(handler=handler, token=token, filters=filters)

        self.subscriptions[subscription.get_subscription_id()] = subscription
        return subscription.get_subscription_id()

    def delete_subscription(self, *, sid: ID, token: AuthToken):
        self.get_subscription(sid=sid, token=token)
        if sid in self.subscriptions:
            self.subscriptions.pop(sid)

    def get_subscription(self, *, sid: ID, token: AuthToken) -> AEventSubscription:
        if not sid in self.subscriptions:
            raise EventException("Invalid subscription")

        subscription = self.subscriptions[id]
        if not subscription.has_access(token=token):
            raise EventException("Access denied")
        return subscription

    def drain_events(self, *, sid: ID, token: AuthToken, timeout: int) -> list:
        subscription = self.get_subscription(sid=sid, token=token)
        return subscription.drain_events(timeout=timeout)

    def dispatch_event(self, *, event: IEvent):
        try:
            for s in self.subscriptions.values():
                if s.is_abandoned():
                    if s.get_subscription_id() in self.subscriptions:
                        self.subscriptions.pop(s.get_subscription_id())
                else:
                    s.deliver_event(event=event)
        except Exception as e:
            print("Could not dispatch event {}".format(e))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)


class Plugin:
    """
    Plugin class describes metadatata about an plugin supplied
    """
    def __init__(self):
        self.id = None
        self.plugin_type = None
        self.factory = False
        self.name = None
        self.description = None
        self.class_name = None
        self.config_properties = None
        self.config_template = None
        self.actor_type = 0

    def is_factory(self) -> bool:
        """
        Checks if this plugin is a factory.
        @return true if the plugin represents a factory
        """
        return self.factory

    def set_factory(self, *, factory: bool):
        """
        Sets the factory flag. A plugin is a factory if it is used to create the
        actual plugin rather than it being the plugin itself.
        @param factory true|false
        """
        self.factory = factory

    def get_id(self) -> str:
        """
        Returns the plugin identifier.
        @return the id
        """
        return self.id

    def set_id(self, *, pid: str):
        """
        @param pid the id to set
        """
        self.id = id

    def get_plugin_type(self) -> int:
        """
        @return the pluginType
        """
        return self.plugin_type

    def set_plugin_type(self, *, plugin_type: int):
        """
        @param plugin_type the pluginType to set
        """
        self.plugin_type = plugin_type

    def get_class_name(self) -> str:
        """
        @return the className
        """
        return self.class_name

    def set_class_name(self, *, class_name: str):
        """
        @param class_name the className to set
        """
        self.class_name = class_name

    def get_config_properties(self) -> dict:
        """
        Get config properties
        @return config properties
        """
        return self.config_properties

    def set_config_properties(self, *, config_properties: dict):
        """
        Set config properties
        @param config_properties config properties
        """
        self.config_properties = config_properties

    def get_config_template(self) -> str:
        """
        Get Config template
        @param config template
        """
        return self.config_template

    def set_config_template(self, *, config_template: str):
        """
        Set config template
        @param config_template config template
        """
        self.config_template = config_template

    def get_name(self) -> str:
        """
        Get Name
        @return name
        """
        return self.name

    def set_name(self, *, name: str):
        """
        Set name
        @param name name
        """
        self.name = name

    def get_actor_type(self) -> int:
        """
        Get Actor Type
        @return actor type
        """
        return self.actor_type

    def set_actor_type(self, *, actor_type: int):
        """
        Set actor type
        @param actor_type actor type
        """
        self.actor_type = actor_type



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import pickle
from typing import TYPE_CHECKING

from fabric.actor.boot.inventory.neo4j_resource_pool_factory import Neo4jResourcePoolFactory
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.apis.i_slice_factory import ISliceFactory
from fabric.actor.core.common.exceptions import SliceException
from fabric.actor.core.kernel.slice import Slice
from fabric.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_slice import ISlice
    from fabric.actor.core.util.resource_data import ResourceData


class SliceFactory(ISliceFactory):
    @staticmethod
    def create(*, slice_id: ID, name: str = None, data: ResourceData = None) -> ISlice:
        return Slice(slice_id=slice_id, name=name, data=data)

    @staticmethod
    def create_instance(*, properties: dict) -> ISlice:
        if Constants.property_pickle_properties not in properties:
            raise SliceException(Constants.invalid_argument)

        serialized_slice = properties[Constants.property_pickle_properties]
        deserialized_slice = pickle.loads(serialized_slice)
        if deserialized_slice.get_graph_id() is not None:
            graph_id = str(deserialized_slice.get_graph_id())
            arm_graph = Neo4jResourcePoolFactory.get_arm_graph(graph_id=graph_id)
            deserialized_slice.set_graph(arm_graph)
        return deserialized_slice



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_actor import IActor
from fabric.actor.core.apis.i_client_public import IClientPublic
from fabric.actor.core.apis.i_delegation import IDelegation

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_broker_proxy import IBrokerProxy
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.actor.core.apis.i_slice import ISlice
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.util.reservation_set import ReservationSet


class IClientActor(IActor, IClientPublic):
    """
    IClientActor defines the common functionality for actors acting
    as clients of other actors (controllers and brokers). Every client actor
    is connected to one or more server actors. Each server actor is represented
    as a proxy object. Client actors maintain a registry of proxies to server
    actors that they are connected to. Proxies to actors acting in the broker
    role are explicitly managed. Proxies for actors acting in the site authority
    role are automatically managed as they are embedded in tickets sent from brokers.
    """

    @abstractmethod
    def add_broker(self, *, broker: IBrokerProxy):
        """
        Registers a broker. If this is the first broker to be registered, it is
        set as the default broker.

        @params broker broker to register
        """

    @abstractmethod
    def demand(self, *, rid: ID):
        """
        Demand a reservation
        @params reservation: reservation
        """

    @abstractmethod
    def extend_ticket_client(self, *, reservation: IClientReservation):
        """
        Issues a ticket extend request for the given reservation. Note: the
        reservation must have already been registered with the actor.

        @param reservation reservation to extend the ticket for
        @param rset set of reservations to extend tickets for

        @throws Exception in case of error
        """

    @abstractmethod
    def extend_tickets_client(self, *, rset: ReservationSet):
        """
        Issues a ticket extend request for the given reservation. Note: the
        reservation must have already been registered with the actor.

        @param reservation reservation to extend the ticket for
        @param rset set of reservations to extend tickets for

        @throws Exception in case of error
        """

    @abstractmethod
    def get_broker(self, *, guid: ID) -> IBrokerProxy:
        """
        Gets the broker proxy with the given guid

        @param guid broker guid

        @return requested broker
        """

    @abstractmethod
    def get_brokers(self) -> list:
        """
        Returns all brokers registered with the actor.

        @return an array of brokers
        """

    @abstractmethod
    def get_default_broker(self) -> IBrokerProxy:
        """
        Returns the default broker.

        @return the default broker
        """

    @abstractmethod
    def ticket_client(self, *, reservation: IClientReservation):
        """
        Issues a ticket request for the given reservation. Note: the reservation
        must have already been registered with the actor.

        All exceptions are caught and logged but no exception is propagated. No information will
        be delivered to indicate that some failure has taken place, e.g., failure
        to communicate with a broker. Inspect the state of individual
        reservations to determine whether/what failures have taken place.

        @param reservation reservation to obtain a ticket for

        @throws Exception in case of error
        """

    @abstractmethod
    def tickets_client(self, *, rset: ReservationSet):
        """
        Issues a ticket request for the given reservation. Note: the reservation
        must have already been registered with the actor.

        All exceptions are caught and logged but no exception is propagated. No information will
        be delivered to indicate that some failure has taken place, e.g., failure
        to communicate with a broker. Inspect the state of individual
        reservations to determine whether/what failures have taken place.

        @param rset set of reservations to obtain tickets for

        @throws Exception in case of error
        """

    @abstractmethod
    def modify(self, *, reservation_id: ID, modify_properties: dict):
        """
        Issue modify request for given reservation. Note: the reservation
        must have already been registered with the actor.

        @param reservation_id reservationID for the reservation to modify
        @param modify_properties property list for modify
        @throws Exception in case of error
        """

    @abstractmethod
    def claim_delegation_client(self, *, delegation_id: str = None, slice_object: ISlice = None,
                                broker: IBrokerProxy = None, id_token: str = None) -> IDelegation:
        """
        Claims already exported resources from the given broker. The delegation
        will be stored in the default slice.

        @param delegation_id delegation identifier of the exported delegation
        @param slice_object slice
        @param broker broker proxy
        @param id_token id token

        @returns delegation
        @raises Exception in case of failure
        """

    @abstractmethod
    def reclaim_delegation_client(self, *, delegation_id: str = None, slice_object: ISlice = None,
                                  broker: IBrokerProxy = None, id_token: str = None) -> IDelegation:
        """
        Reclaims already exported resources from the given broker. The delegation
        will be stored in the default slice.

        @param delegation_id delegation identifier of the exported delegation
        @param slice_object slice
        @param broker broker proxy
        @param id_token id token

        @returns delegation
        @raises Exception in case of failure
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.common.resource_vector import ResourceVector
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_type import ResourceType


class ResourceDelegation:
    """
    ResourceDelegation expresses a delegation of a number of resources with
    identical properties from one actor to another. Each delegation is backed by resources
    stored in one or more resource bins and covers a fixed time interval (term). Individual units
    in a delegation are indistinguishable from each other.
    """
    def __init__(self, *, units: int = None, vector: ResourceVector = None, term: Term = None,
                 rtype: ResourceType = None, sources: list = None, bins: list = None, properties: dict = None,
                 issuer: ID = None, holder: ID = None):
        # The delegation's unique identifier.
        self.guid = ID()
        # Lease interval.
        self.term = term
        # Number of units delegated. This is the total number of
        # units represented by the delegation and can be different
        # from the number of physical units
        self.units = units
        # Resource vector for each delegated unit.
        self.vector = vector
        # Resource type.
        self.type = rtype
        # Resource properties (optional).
        self.properties = properties
        # Issuer identifier.
        self.issuer = issuer
        # Holder identifier.
        self.holder = holder
        # Source bins used for this delegation.
        self.sources = sources
        # Actor-local resource bins referenced by the delegation.
        self.bins = bins

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.vector = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['vector']
        return state

    def __str__(self):
        result = "delegation=[guid={},units={},vector={},type={},issuer={},holder={}]".format(self.guid, self.units,
                                                                                              self.vector, self.type,
                                                                                              self.issuer, self.holder)
        return result

    def is_valid(self) -> bool:
        if self.guid is None or self.term is None or self.units <= 0 or self.vector is None or \
                not self.vector.is_positive() or self.type is None or self.sources is None \
                or len(self.sources) == 0 or self.bins is None or len(self.bins) == 0:
            return False
        return True

    def get_guid(self) -> ID:
        """
        Get guid
        @return guid
        """
        return self.guid

    def get_resource_vector(self) -> ResourceVector:
        """
        Get Resource Vector
        @return resource vector
        """
        return self.vector

    def get_resource_type(self) -> ResourceType:
        """
        Get Resource Type
        @return resource type
        """
        return self.type

    def get_properties(self) -> dict:
        """
        Get Properties
        @return properties
        """
        return self.properties

    def get_term(self) -> Term:
        """
        Get Term
        @return term
        """
        return self.term

    def get_units(self) -> int:
        """
        Get Units
        @return units
        """
        return self.units

    def get_issuer(self) -> ID:
        """
        Get Issuer
        @return issuer
        """
        return self.issuer

    def get_holder(self) -> ID:
        """
        Get holder
        @return holder
        """
        return self.holder

    def set_issuer(self, *, guid: ID):
        """
        Set issuer
        @param guid issuer guid
        """
        self.issuer = guid

    def set_holder(self, *, guid: ID):
        """
        Set holder
        @param guid holder guid
        """
        self.holder = guid

    def set_units(self, *, units: int):
        """
        Set units
        @param units units
        """
        self.units = units

    def get_map(self):
        """
        Returns a hash map of all actor-local bins referenced by this delegation.
        The map indexes bins by guid.
        @return hash map
        """
        result = {}
        for i in range(len(self.bins)):
            result[self.bins[i].guid] = self.bins[i]
        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, List

from fabric.actor.core.apis.i_component import IComponent
from fabric.actor.core.manage.messages.event_mng import EventMng
from fabric.message_bus.messages.delegation_avro import DelegationAvro

if TYPE_CHECKING:
    from fabric.actor.core.util.id import ID
    from fabric.message_bus.messages.reservation_mng import ReservationMng
    from fabric.message_bus.messages.reservation_state_avro import ReservationStateAvro
    from fabric.message_bus.messages.slice_avro import SliceAvro


class IMgmtActor(IComponent):
    @abstractmethod
    def get_slices(self, *, id_token: str = None, slice_id: ID = None) -> List[SliceAvro]:
        """
        Obtains all slices.
        @param id_token id token
        @param slice_id slice id
        @return returns list of all the slices
        """

    @abstractmethod
    def add_slice(self, *, slice_obj: SliceAvro) -> ID:
        """
        Adds a new slice
        @param slice_obj slice
        @return returns slice id
        """

    @abstractmethod
    def remove_slice(self, *, slice_id: ID) -> bool:
        """
        Removes the specified slice
        @param slice_id slice id
        @return true for success; false otherwise
        """

    @abstractmethod
    def update_slice(self, *, slice_obj: SliceAvro) -> bool:
        """
        Updates the specified slice.
        The only updatable slice attributes are:
        - description
        - all properties lists
        @param slice_obj slice
        @return true for success; false otherwise
        """

    @abstractmethod
    def get_reservations(self, *, id_token: str = None, state: int = None,
                         slice_id: ID = None, rid: ID = None) -> List[ReservationMng]:
        """
        @param id_token id token
        @param state state
        @param slice_id slice ID
        @param rid reservation id
        Obtains all reservations
        @return returns list of the reservations
        """

    @abstractmethod
    def update_reservation(self, *, reservation: ReservationMng) -> bool:
        """
        Updates the specified reservation
        The fields that can be updated are:

        - all properties lists

        @param reservation reservation to be updated
        @return true for success; false otherwise
        """

    @abstractmethod
    def close_reservation(self, *, rid: ID) -> bool:
        """
        Closes the specified reservation
        @param rid reservation id
        @return true for success; false otherwise
        """

    @abstractmethod
    def close_reservations(self, *, slice_id: ID) -> bool:
        """
        Closes all reservations in the specified slice.
        @param slice_id slice ID
        @return true for success; false otherwise
        """

    @abstractmethod
    def remove_reservation(self, *, rid: ID) -> bool:
        """
        Removes the specified reservation.
        Note only closed reservations can be removed.
        @param rid reservation id of the reservation to be removed
        @return true for success; false otherwise
        """

    @abstractmethod
    def get_reservation_state_for_reservations(self, *, reservation_list: List[ID],
                                               id_token: str = None) -> List[ReservationStateAvro]:
        """
        Returns the state of each of the specified reservations.
        The order in the return list matches the order in the @reservations li
        @param reservation_list list of reservations
        @param id_token id token
        @return list of state of the specified reservations
        """

    @abstractmethod
    def get_name(self) -> str:
        """
        Returns the name of the actor.
        @return returns name of the actor
        """

    @abstractmethod
    def get_guid(self) -> ID:
        """
        Returns the guid of the actor.
        @return returns guid of the actor
        """

    @abstractmethod
    def clone(self):
        """
        Creates clone of this proxy to make it possible to be used by a separate thread.
        @return returns the cloned actor
        """

    @abstractmethod
    def get_delegations(self, *, slice_id: ID = None, state: int = None,
                        delegation_id: ID = None, id_token: str = None) -> List[DelegationAvro]:
        """
        Get Delegations
        @param slice_id slice id
        @param state state
        @param delegation_id delegation id
        @param id_token id token
        @return returns list of the delegations
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_authority_proxy import IAuthorityProxy
from fabric.actor.core.apis.i_controller_callback_proxy import IControllerCallbackProxy
from fabric.actor.core.apis.i_controller_reservation import IControllerReservation
from fabric.actor.core.apis.i_rpc_request_state import IRPCRequestState
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import ProxyException
from fabric.actor.core.kernel.authority_reservation_factory import AuthorityReservationFactory
from fabric.actor.core.proxies.local.local_broker import LocalBroker
from fabric.actor.core.proxies.local.local_proxy import LocalProxy
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.security.auth_token import AuthToken


class LocalAuthority(LocalBroker, IAuthorityProxy):
    def _prepare(self, *, reservation: IControllerReservation, callback: IControllerCallbackProxy,
                 caller: AuthToken) -> IRPCRequestState:
        state = LocalProxy.LocalProxyRequestState()
        state.reservation = self.pass_reservation_authority(reservation=reservation, auth=caller)
        state.callback = callback
        return state

    def prepare_redeem(self, *, reservation: IControllerReservation, callback: IControllerCallbackProxy,
                       caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, callback=callback, caller=caller)

    def prepare_extend_lease(self, *, reservation: IControllerReservation, callback: IControllerCallbackProxy,
                             caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, callback=callback, caller=caller)

    def prepare_modify_lease(self, *, reservation: IControllerReservation, callback: IControllerCallbackProxy,
                             caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, callback=callback, caller=caller)

    def prepare_close(self, *, reservation: IControllerReservation, callback: IControllerCallbackProxy,
                      caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, callback=callback, caller=caller)

    def pass_reservation_authority(self, *, reservation: IControllerReservation, auth: AuthToken) -> IReservation:
        if reservation.get_resources().get_resources() is None:
            raise ProxyException(Constants.not_specified_prefix.format("concrete set"))

        slice_obj = reservation.get_slice().clone_request()
        term = reservation.get_term().clone()

        rset = self.abstract_clone_authority(rset=reservation.get_resources())
        rset.get_resource_data().configuration_properties = ResourceData.merge_properties(
            from_props=reservation.get_slice().get_config_properties(),
            to_props=rset.get_resource_data().get_configuration_properties())

        original_ticket = reservation.get_resources().get_resources()
        encoded_ticket = original_ticket.encode(protocol=Constants.protocol_local)
        from fabric.actor.core.proxies.proxy import Proxy
        decoded_ticket = Proxy.decode(encoded=encoded_ticket, plugin=self.get_actor().get_plugin())
        rset.set_resources(cset=decoded_ticket)

        authority_reservation = AuthorityReservationFactory.create(resources=rset, term=term, slice_obj=slice_obj,
                                                                   rid=reservation.get_reservation_id())
        authority_reservation.set_sequence_in(sequence=reservation.get_lease_sequence_out())
        authority_reservation.set_owner(owner=self.get_identity())

        return authority_reservation



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_authority_reservation import IAuthorityReservation
from fabric.actor.core.apis.i_callback_proxy import ICallbackProxy
from fabric.actor.core.apis.i_controller_callback_proxy import IControllerCallbackProxy
from fabric.actor.core.apis.i_rpc_request_state import IRPCRequestState
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.kernel.rpc_request_type import RPCRequestType
from fabric.actor.core.proxies.local.local_return import LocalReturn
from fabric.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.update_data import UpdateData
from fabric.actor.security.auth_token import AuthToken
from fabric.actor.test.client_callback_helper import ClientCallbackHelper
from fabric.actor.test.test_exception import TestException


class ControllerCallbackHelper(ClientCallbackHelper, IControllerCallbackProxy):
    class IUpdateLeaseHandler:
        def handle_update_lease(self, *, reservation: IReservation, update_data: UpdateData, caller: AuthToken):
            """
            Implemented by derived class
            """

        def check_termination(self):
            """
            Implemented by derived class
            """

    def __init__(self, *, name: str, guid: ID):
        super().__init__(name=name, guid=guid)
        self.called_for_lease = 0
        self.lease = None
        self.updateLeaseHandler = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['updateLeaseHandler']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.updateLeaseHandler = None

    def prepare_update_lease(self, *, reservation: IAuthorityReservation, update_data: UpdateData,
                             callback: ICallbackProxy, caller: AuthToken) -> IRPCRequestState:
        state = ClientCallbackHelper.MyRequestState()
        state.reservation = LocalReturn.pass_reservation(reservation=reservation,
                                                         plugin=ActorRegistrySingleton.get().get_actor(
                                                             actor_name_or_guid=self.token.get_name()).get_plugin())
        state.update_data = UpdateData()
        state.update_data.absorb(other=update_data)
        return state

    def execute(self, *, request: IRPCRequestState):
        if request.get_type() == RPCRequestType.UpdateLease:
            self.lease = request.reservation
            self.called_for_lease += 1
            if self.updateLeaseHandler is not None:
                self.updateLeaseHandler.handle_update_lease(reservation=self.lease, update_data=request.update_data,
                                                            caller=request.get_caller())
            else:
                super().execute(request=request)

    def get_lease(self) -> IReservation:
        return self.lease

    def get_called_for_lease(self):
        return self.called_for_lease

    def set_update_lease_handler(self, *, handler: IUpdateLeaseHandler):
        self.updateLeaseHandler = handler

    def get_called(self) -> int:
        return self.called

    def get_guid(self) -> ID:
        return self.token.get_guid()

    def get_identity(self) -> AuthToken:
        return self.token

    def get_name(self) -> str:
        return self.token.get_name()

    def get_reservation(self) -> IReservation:
        return self.reservation

    def get_type(self):
        return Constants.protocol_local

    def prepare_query_result(self, *, request_id: str, response, caller: AuthToken) -> IRPCRequestState:
        raise TestException(Constants.not_implemented)

    def prepare_failed_request(self, *, request_id: str, failed_request_type,
                               failed_reservation_id: ID, error: str, caller: AuthToken) -> IRPCRequestState:
        raise TestException(Constants.not_implemented)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import json
import traceback

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from fabric.actor.boot.inventory.neo4j_resource_pool_factory import Neo4jResourcePoolFactory
from fabric.actor.core.apis.i_actor import ActorType
from fabric.actor.core.apis.i_mgmt_actor import IMgmtActor
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.util.id import ID
from fabric.actor.security.fabric_token import FabricToken
from fabric.actor.security.pdp_auth import PdpAuth, ActionId, ResourceType
from fabric.orchestrator.core.orchestrator_state import OrchestratorStateSingleton
from fim.graph.neo4j_property_graph import Neo4jPropertyGraph


class OrchestratorHandler:
    def __init__(self):
        self.controller_state = OrchestratorStateSingleton.get()
        from fabric.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()
        self.token_public_key = GlobalsSingleton.get().get_config().get_oauth_config().get(
            Constants.property_conf_o_auth_token_public_key, None)
        self.pdp_config = GlobalsSingleton.get().get_config().get_global_config().get_pdp_config()

    def get_logger(self):
        return self.logger

    def validate_credentials(self, *, token) -> dict:
        try:
            fabric_token = FabricToken(token_public_key=self.token_public_key, logger=self.logger,
                                            token=token)

            return fabric_token.validate()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("Exception occurred while validating the token e: {}".format(e))

    def check_access(self, *, action_id: ActionId, resource_type: ResourceType, token: str,
                     resource_id: str = None) -> bool:
        fabric_token = FabricToken(token_public_key=self.token_public_key, logger=self.logger,
                                   token=token)
        fabric_token.validate()
        pdp_auth = PdpAuth(config=self.pdp_config, logger=self.logger)
        return pdp_auth.check_access(fabric_token=fabric_token.get_decoded_token(),
                                     actor_type=ActorType.Orchestrator,
                                     action_id=action_id, resource_type=resource_type,
                                     resource_id=resource_id)

    def get_broker(self, *, controller: IMgmtActor) -> ID:
        try:
            brokers = controller.get_brokers()
            if brokers is not None:
                return ID(uid=next(iter(brokers), None).get_guid())

        except Exception as e:
            self.logger.debug(traceback.format_exc())

        return None

    def discover_types(self, *, controller: IMgmtActor, token: str) -> dict:
        broker = self.get_broker(controller=controller)
        if broker is None:
            raise Exception("Unable to determine broker proxy for this controller. "
                            "Please check Orchestrator container configuration and logs.")

        self.controller_state.set_broker(broker=str(broker))

        my_pools = controller.get_pool_info(broker=broker, id_token=token)
        if my_pools is None:
            raise Exception("Could not discover types: {}".format(controller.get_last_error()))

        response = None
        for p in my_pools:
            try:
                bqm = p.properties.get(Constants.broker_query_model, None)
                if bqm is not None:
                    graph = Neo4jResourcePoolFactory.get_graph_from_string(graph_str=bqm)
                    graph.validate_graph()
                    Neo4jResourcePoolFactory.delete_graph(graph_id=graph.get_graph_id())
                    response = bqm
            except Exception as e:
                self.logger.error(traceback.format_exc())
                self.logger.debug("Could not process discover types response {}".format(e))

        return response

    def list_resources(self, *, token: str):
        try:
            self.check_access(action_id=ActionId.query, resource_type=ResourceType.resources, token=token)
            self.controller_state.close_dead_slices()
            controller = self.controller_state.get_management_actor()
            self.logger.debug("list resources invoked controller:{}".format(controller))

            try:
                abstract_models = self.discover_types(controller=controller, token=token)
            except Exception as e:
                self.logger.error("Failed to populate abstract models e: {}".format(e))
                raise e

            if abstract_models is None:
                raise Exception("Failed to populate abstract models")

            return abstract_models

        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("Exception occurred processing list resource e: {}".format(e))
            raise e



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import threading
from typing import TYPE_CHECKING

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import PluginException

if TYPE_CHECKING:
    from fabric.actor.core.plugins.base_plugin import BasePlugin
    from fabric.actor.core.plugins.config.config_token import ConfigToken
    from fabric.actor.core.plugins.config.configuration_mapping import ConfigurationMapping


class Config:
    # Specifies a configuration file or action to execute.
    property_config = "config.action"
    property_exception_message = "exception.message"
    property_exception_stack = "exception.stack"
    property_target_name = "target.name"
    property_target_result_code = "target.code"
    property_target_result_code_message = "target.code.message"
    property_configuration_properties = "config"
    property_action_sequence_number = "action.sequence"

    result_code_exception = -1
    result_code_ok = 0
    target_create = "create"
    target_delete = "delete"
    target_modify = "modify"
    property_resource_type = "unit.resourceType"
    property_unit_all = "unit.all"

    def __init__(self):
        self.is_sync = False
        self.plugin = None
        self.logger = None
        self.initialized = False
        self.config_mappings = {}
        self.lock = threading.Lock()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['plugin']
        del state['initialized']
        del state['lock']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None
        self.plugin = None
        self.initialized = False
        self.lock = threading.Lock()

    def initialize(self):
        if not self.initialized:
            if self.plugin is None:
                raise PluginException(Constants.not_specified_prefix.format("plugin"))
            self.logger = self.plugin.get_logger()
            self.initialized = True

    def add_config_mapping(self, *, mapping: ConfigurationMapping):
        try:
            self.lock.acquire()
            self.config_mappings[mapping.get_key()] = mapping
        finally:
            self.lock.release()

    def create(self, *, token: ConfigToken, properties: dict):
        self.logger.info("Executing Join")

        result = {self.property_target_name: self.target_create,
                  self.property_target_result_code: self.result_code_ok,
                  self.property_action_sequence_number: 0}

        self.plugin.configuration_complete(token=token, properties=result)
        self.logger.info("Executing Join completed")

    def delete(self, *, token: ConfigToken, properties: dict):
        self.logger.info("Executing Leave")

        result = {self.property_target_name: self.target_delete,
                  self.property_target_result_code: self.result_code_ok,
                  self.property_action_sequence_number: 0}

        self.plugin.configuration_complete(token=token, properties=result)
        self.logger.info("Executing Leave completed")

    def modify(self, *, token: ConfigToken, properties: dict):
        self.logger.info("Executing Modify")

        result = {self.property_target_name: self.target_modify,
                  self.property_target_result_code: self.result_code_ok,
                  self.property_action_sequence_number: 0}

        self.plugin.configuration_complete(token=token, properties=result)
        self.logger.info("Executing Modify completed")

    def set_logger(self, *, logger):
        self.logger = logger

    def set_slices_plugin(self, *, plugin: BasePlugin):
        self.plugin = plugin

    @staticmethod
    def get_action_sequence_number(*, properties: dict):
        if properties is None:
            raise PluginException(Constants.invalid_argument)

        if Config.property_action_sequence_number in properties:
            return int(properties[Config.property_action_sequence_number])

        raise PluginException("Action Sequence Number not found")

    @staticmethod
    def get_result_code(*, properties: dict):
        if properties is None:
            raise PluginException(Constants.invalid_argument)

        if Config.property_target_result_code in properties:
            return int(properties[Config.property_target_result_code])

        raise PluginException("Target Result code not found")

    @staticmethod
    def get_result_code_message(*, properties: dict):
        if properties is None:
            raise PluginException(Constants.invalid_argument)

        if Config.property_target_result_code_message in properties:
            return properties[Config.property_target_result_code_message]

        return None

    @staticmethod
    def get_exception_message(*, properties: dict):
        if properties is None:
            raise PluginException(Constants.invalid_argument)

        if Config.property_exception_message in properties:
            return properties[Config.property_exception_message]

        return None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import List

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import ManageException
from fabric.actor.core.manage.container_management_object import ContainerManagementObject
from fabric.actor.core.manage.management_object import ManagementObject
from fabric.actor.core.apis.i_mgmt_actor import IMgmtActor
from fabric.actor.core.apis.i_mgmt_authority import IMgmtAuthority
from fabric.actor.core.apis.i_mgmt_broker import IMgmtBroker
from fabric.actor.core.apis.i_mgmt_container import IMgmtContainer
from fabric.actor.core.apis.i_mgmt_controller import IMgmtController
from fabric.actor.core.manage.local.local_proxy import LocalProxy
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.reflection_utils import ReflectionUtils
from fabric.actor.security.auth_token import AuthToken
from fabric.message_bus.messages.actor_avro import ActorAvro
from fabric.message_bus.messages.proxy_avro import ProxyAvro


class LocalContainer(LocalProxy, IMgmtContainer):
    def __init__(self, *, manager: ManagementObject, auth: AuthToken):
        super().__init__(manager=manager, auth=auth)
        if not isinstance(manager, ContainerManagementObject):
            raise ManageException("Invalid manager object. Required: {}".format(type(ContainerManagementObject)))

    def get_management_object(self, *, key: ID):
        try:
            obj = self.manager.get_management_object(key=key)

            if obj is None:
                return None

            desc_list = obj.get_proxies()

            if desc_list is None:
                raise ManageException("Management object did not specify any proxies")
            desc = None
            for d in desc_list:
                if d.get_protocol() == Constants.protocol_local:
                    desc = d
                    break

            if desc is None or desc.get_proxy_class() is None or desc.get_proxy_module() is None:
                raise ManageException("Manager object did not specify local proxy")

            try:
                mgmt_obj = ReflectionUtils.create_instance_with_params(module_name=desc.get_proxy_module(),
                                                                       class_name=desc.get_proxy_class())(manager=obj,
                                                                                                          auth=self.auth)
                return mgmt_obj
            except Exception as e:
                traceback.print_exc()
                raise ManageException("Could not instantiate proxy {}".format(e))
        except Exception as e:
            traceback.print_exc()
            raise e

    def get_actor(self, *, guid: ID):
        try:
            component = self.get_management_object(key=guid)

            if component is not None:
                if isinstance(component, IMgmtActor):
                    return component
                else:
                    self.last_exception = Exception(Constants.invalid_management_object_type.format(type(component)))

            return None
        except Exception as e:
            self.last_exception = e
            traceback.print_exc()

    def get_actors(self) -> List[ActorAvro]:
        self.clear_last()
        try:
            result = self.manager.get_actors(caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return result.result

        except Exception as e:
            self.last_exception = e
            traceback.print_exc()

        return None

    def get_actors_from_database(self) -> List[ActorAvro]:
        self.clear_last()
        try:
            result = self.manager.get_actors_from_database(caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return result.result

        except Exception as e:
            self.last_exception = e
            traceback.print_exc()

        return None

    def get_actors_from_database_name_type_status(self, *, name: str, actor_type: int, status: int) -> List[ActorAvro]:
        self.clear_last()
        try:
            result = self.manager.get_actors_from_database_name_type_status(name=name, actor_type=actor_type,
                                                                            status=status, caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return result.result

        except Exception as e:
            self.last_exception = e
            traceback.print_exc()

        return None

    def get_authorities(self) -> List[ActorAvro]:
        self.clear_last()
        try:
            result = self.manager.get_authorities(caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return result.result

        except Exception as e:
            self.last_exception = e
            traceback.print_exc()

        return None

    def get_brokers(self) -> List[ActorAvro]:
        self.clear_last()
        try:
            result = self.manager.get_brokers(caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return result.result

        except Exception as e:
            self.last_exception = e
            traceback.print_exc()

        return None

    def get_controllers(self) -> List[ActorAvro]:
        self.clear_last()
        try:
            result = self.manager.get_controllers(caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return result.result

        except Exception as e:
            self.last_exception = e
            traceback.print_exc()

        return None

    def get_proxies(self, *, protocol: str) -> List[ProxyAvro]:
        self.clear_last()
        try:
            result = self.manager.get_proxies_by_protocol(protocol=protocol, caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return result.result

        except Exception as e:
            self.last_exception = e
            traceback.print_exc()

        return None

    def get_broker_proxies(self, *, protocol: str) -> List[ProxyAvro]:
        self.clear_last()
        try:
            result = self.manager.get_broker_proxies(protocol=protocol, caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return result.result

        except Exception as e:
            self.last_exception = e
            traceback.print_exc()

        return None

    def get_authority_proxies(self, *, protocol: str) -> List[ProxyAvro]:
        self.clear_last()
        try:
            result = self.manager.get_site_proxies(protocol=protocol, caller=self.auth)
            self.last_status = result.status

            if result.status.get_code() == 0:
                return result.result

        except Exception as e:
            self.last_exception = e
            traceback.print_exc()

        return None

    def get_controller(self, *, guid: ID) -> IMgmtController:
        component = self.get_management_object(key=guid)

        if component is None:
            return None

        self.clear_last()

        if not isinstance(component, IMgmtController):
            self.last_exception = Exception("Invalid type")
        else:
            return component

        return None

    def get_broker(self, *, guid: ID) -> IMgmtBroker:
        component = self.get_management_object(key=guid)

        if component is not None:
            if isinstance(component, IMgmtBroker):
                return component
            else:
                self.last_exception = Exception(Constants.invalid_management_object_type.format(type(component)))

        return None

    def get_authority(self, *, guid: ID) -> IMgmtAuthority:
        component = self.get_management_object(key=guid)

        if component is not None:
            if isinstance(component, IMgmtAuthority):
                return component
            else:
                self.last_exception = Exception(Constants.invalid_management_object_type.format(type(component)))

        return None


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.time.actor_clock import ActorClock
from fabric.actor.core.time.calendar.client_calendar import ClientCalendar
from fabric.actor.core.util.reservation_list import ReservationList
from fabric.actor.core.util.reservation_set import ReservationSet


class ControllerCalendar(ClientCalendar):
    """
    Controller calendar. In addition to the lists maintained by ClientCalendar,
    this class maintains the following lists:
     - closing: a list of reservations organized by the cycle they must be closed
     - redeeming: a list of reservations organized by the cycle they must be redeemed
    """
    def __init__(self, *, clock: ActorClock):
        """
        Constructor
        @params clock: clock
        """
        super().__init__(clock=clock)
        self.closing = ReservationList()
        self.redeeming = ReservationList()

    def remove(self, *, reservation: IReservation):
        """
        Removes the reservation from the calendar.

        @params reservation: reservation to remove
        """
        super().remove(reservation=reservation)
        self.remove_closing(reservation=reservation)
        self.remove_redeeming(reservation=reservation)

    def get_closing(self, *, cycle: int) -> ReservationSet:
        """
        Returns all reservations that need to be closed up to and including the specified cycle.

        @params cycle : cycle

        @returns a set of reservations that must be closed on the specified cycle
        """
        try:
            self.lock.acquire()
            return self.closing.get_all_reservations(cycle=cycle)
        finally:
            self.lock.release()

    def add_closing(self, *, reservation: IReservation, cycle: int):
        """
        Adds a reservation to be closed on the specified cycle.

        @params reservation : reservation to add
        @params cycle : cycle
        """
        try:
            self.lock.acquire()
            self.closing.add_reservation(reservation=reservation, cycle=cycle)
        finally:
            self.lock.release()

    def remove_closing(self, *, reservation: IReservation):
        """
         Removes the given reservation from the closing list.

        @params reservation : reservation to remove
        """
        try:
            self.lock.acquire()
            self.closing.remove_reservation(reservation=reservation)
        finally:
            self.lock.release()

    def get_redeeming(self, *, cycle: int) -> ReservationSet:
        """
        Returns all reservations that need to be redeemed up to and including the specified cycle.

        @params cycle : cycle

        @returns a set of reservations that must be redeemed on the specified cycle
        """
        try:
            self.lock.acquire()
            return self.redeeming.get_all_reservations(cycle=cycle)
        finally:
            self.lock.release()

    def add_redeeming(self, *, reservation: IReservation, cycle: int):
        """
        Adds a reservation to be redeeming on the specified cycle.

        @params reservation : reservation to add
        @params cycle : cycle
        """
        try:
            self.lock.acquire()
            self.redeeming.add_reservation(reservation=reservation, cycle=cycle)
        finally:
            self.lock.release()

    def remove_redeeming(self, *, reservation: IReservation):
        """
         Removes the given reservation from the redeeming list.

        @params reservation : reservation to remove
        """
        try:
            self.lock.acquire()
            self.redeeming.remove_reservation(reservation=reservation)
        finally:
            self.lock.release()

    def tick(self, *, cycle: int):
        super().tick(cycle=cycle)
        try:
            self.lock.acquire()
            self.closing.tick(cycle=cycle)
            self.redeeming.tick(cycle=cycle)
        finally:
            self.lock.release()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import PolicyException
from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor
from fabric.actor.core.core.unit import Unit, UnitState
from fabric.actor.core.core.unit_set import UnitSet
from fabric.actor.core.policy.resource_control import ResourceControl
from fabric.actor.core.policy.vm_control import VMControl
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.ipv4_set import IPv4Set
from fabric.actor.core.kernel.resource_set import ResourceSet
if TYPE_CHECKING:
    from fabric.actor.core.util.resource_type import ResourceType
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.actor.core.apis.i_authority_reservation import IAuthorityReservation
    from fabric.actor.core.apis.i_reservation import IReservation


class PoolData:
    def __init__(self, *, rtype: ResourceType, properties: dict):
        self.total = 0
        self.free_ = 0
        self.rtype = rtype
        self.pd = ResourcePoolDescriptor()
        self.pd.reset(properties=properties, prefix=None)

    def add_units(self, *, count: int):
        self.total += count
        self.free_ += count

    def _allocate_reserve(self, *, count: int):
        if self.free_ < count:
            raise PolicyException("insufficient units (allocate): needed= {} available: {}".format(count, self.free_))

        self.free_ -= count

    def allocate(self, *, count: int):
        self._allocate_reserve(count=count)

    def free(self, *, count: int):
        if self.free_ + count > self.total:
            raise PolicyException("too many units to free")

        self.free_ += count

    def reserve(self, *, count: int):
        self._allocate_reserve(count=count)

    def get_free(self) -> int:
        return self.free_

    def get_total(self) -> int:
        return self.total

    def get_allocated(self) -> int:
        return self.total - self.free_

    def get_descriptor(self) -> ResourcePoolDescriptor:
        return self.pd

    def get_type(self):
        return self.rtype


class SimpleVMControl(ResourceControl):
    def __init__(self):
        super().__init__()
        self.inventory = {}
        self.ipset = IPv4Set()
        self.subnet = None
        self.gateway = None
        self.data_subnet = None
        self.use_ip_set = False

    def __getstate__(self):
        state = self.__dict__.copy()

        del state['authority']
        del state['logger']
        del state['initialized']

        del state['inventory']
        del state['ipset']
        del state['subnet']
        del state['gateway']
        del state['data_subnet']
        del state['use_ip_set']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.authority = None
        self.logger = None
        self.initialized = False

        self.inventory = {}
        self.ipset = IPv4Set()
        self.subnet = None
        self.gateway = None
        self.data_subnet = None
        self.use_ip_set = False

    def donate_reservation(self, *, reservation: IClientReservation):
        traceback.print_stack()
        rset = reservation.get_resources()
        rtype = rset.get_type()
        resource = rset.get_resource_properties()
        local = rset.get_local_properties()

        pool = self.inventory.get(rtype, None)
        if pool is None:
            pool = PoolData(rtype=rtype, properties=resource)
            pool.add_units(count=rset.get_units())
            if VMControl.PropertyIPSubnet in local:
                self.subnet = local[VMControl.PropertyIPSubnet]
            if VMControl.PropertyIPGateway in local:
                self.gateway = local[VMControl.PropertyIPGateway]
            if VMControl.PropertyDataSubnet in local:
                self.data_subnet = local[VMControl.PropertyDataSubnet]
            if VMControl.PropertyIPList in local:
                temp = local[VMControl.PropertyIPList]
                self.ipset.add(ip_list=temp)
                self.use_ip_set = True
            self.inventory[rtype] = pool
        else:
            pool.add_units(count=rset.get_units())

    def assign(self, *, reservation: IAuthorityReservation) -> ResourceSet:
        reservation.set_send_with_deficit(value=True)
        if len(self.inventory) == 0:
            raise PolicyException("no inventory")
        requested = reservation.get_requested_resources()
        request_properties = requested.get_resource_properties()
        rtype = requested.get_type()
        current = reservation.get_resources()

        ticket = requested.get_resources()

        gained = None
        lost = None
        if current is None:
            pool = self.inventory.get(rtype, None)
            if pool is None:
                raise PolicyException(Constants.no_pool)

            needed = ticket.get_units()
            gained = self.get_vms(pool=pool, needed=needed)
            if gained is None or gained.get_units() == 0:
                self.logger.warning("Could not allocate any units for r: {}".format(reservation.get_reservation_id()))
                return None
        else:
            rtype = current.get_type()
            pool = self.inventory.get(rtype, None)
            if pool is None:
                raise PolicyException(Constants.no_pool)
            current_units = current.get_units()
            difference = ticket.get_units() - current_units
            if difference > 0:
                gained = self.get_vms(pool=pool, needed=difference)
            elif difference < 0:
                uset = current.get_resources()
                victims = request_properties[Constants.config_victims]
                to_take = uset.select_extract(count=-difference, victims=victims)
                lost = UnitSet(plugin=self.authority.get_plugin(), units=to_take)
        return ResourceSet(gained=gained, lost=lost, rtype=rtype)

    def get_vms(self, *, pool: PoolData, needed: int) -> UnitSet:
        uset = UnitSet(plugin=self.authority.get_plugin())
        available = min(needed, pool.get_free())

        if self.use_ip_set:
            available = min(available, self.ipset.get_free_count())

            pool.allocate(count=available)

            self.logger.debug("Allocated {} units".format(available))

            for i in range(available):
                vm = Unit(id=ID())
                vm.set_resource_type(rtype=pool.get_type())

                if self.use_ip_set:
                    vm.set_property(name=Constants.unit_management_ip, value=self.ipset.allocate())

                if self.subnet is not None:
                    vm.set_property(name=Constants.unit_manage_subnet, value=self.subnet)

                if self.data_subnet is not None:
                    vm.set_property(name=Constants.unit_data_subnet, value=self.data_subnet)

                if self.gateway is not None:
                    vm.set_property(name=Constants.unit_manage_gateway, value=self.gateway)

                for att in pool.get_descriptor().get_attributes():
                    if att.get_value() is not None:
                        key = att.get_key()
                        key = key.replace("resource.", "unit.")
                        vm.set_property(name=key, value=att.get_value())

                uset.add_unit(u=vm)
        return uset

    def free(self, *, uset: dict):
        if uset is not None:
            for u in uset.values():
                try:
                    self.logger.debug("Freeing 1 unit")
                    rtype = u.get_resource_type()
                    pool = self.inventory.get(rtype, None)
                    if pool is None:
                        raise PolicyException(Constants.no_pool)
                    pool.free(count=1)
                    if self.use_ip_set:
                        self.ipset.free(ip=u.get_property(name=Constants.unit_management_ip))
                except Exception as e:
                    self.logger.error("Failed to release vm {}".format(e))

    def revisit(self, *, reservation: IReservation):
        unit_set = reservation.get_resources().get_resources()
        for u in unit_set.get_set().values():
            try:
                if u.get_state() == UnitState.DEFAULT or \
                        u.get_state() == UnitState.FAILED or \
                        u.get_state() == UnitState.CLOSING or \
                        u.get_state() == UnitState.PRIMING or \
                        u.get_state() == UnitState.ACTIVE or \
                        u.get_state() == UnitState.MODIFYING:
                    rtype = u.get_resource_type()
                    pool = self.inventory.get(rtype, None)
                    if pool is None:
                        raise PolicyException(Constants.no_pool)
                    pool.reserve(1)
                    mgmt_ip = u.get_property(name=Constants.unit_management_ip)
                    if mgmt_ip is not None:
                        self.ipset.reserve(ip=mgmt_ip)
            except Exception as e:
                self.fail(u=u, message="revisit with simplemcontrol", e=e)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from fabric.actor.core.apis.i_reservation import ReservationCategory
from fabric.actor.core.common.constants import ErrorCodes
from fabric.actor.core.kernel.slice import SliceTypes
from fabric.actor.core.manage.kafka.services.kafka_actor_service import KafkaActorService
from fabric.actor.core.manage.management_object import ManagementObject
from fabric.actor.core.proxies.kafka.translate import Translate
from fabric.actor.core.util.id import ID
from fabric.message_bus.messages.add_slice_avro import AddSliceAvro
from fabric.message_bus.messages.get_reservations_request_avro import GetReservationsRequestAvro
from fabric.message_bus.messages.get_slices_request_avro import GetSlicesRequestAvro
from fabric.message_bus.messages.message import IMessageAvro
from fabric.message_bus.messages.result_avro import ResultAvro
from fabric.message_bus.messages.result_reservation_avro import ResultReservationAvro
from fabric.message_bus.messages.result_slice_avro import ResultSliceAvro
from fabric.message_bus.messages.result_string_avro import ResultStringAvro


class KafkaServerActorService(KafkaActorService):
    def __init__(self):
        super().__init__()

    def process(self, *, message: IMessageAvro):
        callback_topic = message.get_callback_topic()
        result = None

        self.logger.debug("Processing message: {}".format(message.get_message_name()))

        if message.get_message_name() == IMessageAvro.GetReservationsRequest and \
                message.get_reservation_type() is not None and \
                message.get_reservation_type() == ReservationCategory.Broker.name:
            result = self.get_reservations_by_category(request=message, category=ReservationCategory.Broker)

        elif message.get_message_name() == IMessageAvro.GetSlicesRequest and \
                message.get_slice_type() is not None and \
                message.get_slice_type() == SliceTypes.InventorySlice.name:
            result = self.get_slices_by_slice_type(request=message, slice_type=SliceTypes.InventorySlice)

        elif message.get_message_name() == IMessageAvro.GetReservationsRequest and \
                message.get_reservation_type() is not None and \
                message.get_reservation_type() == ReservationCategory.Client.name:
            result = self.get_reservations_by_category(request=message, category=ReservationCategory.Inventory)

        elif message.get_message_name() == IMessageAvro.GetSlicesRequest and \
                message.get_slice_type() is not None and \
                message.get_slice_type() == SliceTypes.ClientSlice.name:
            result = self.get_slices_by_slice_type(request=message, slice_type=SliceTypes.ClientSlice)

        elif message.get_message_name() == IMessageAvro.AddSlice and message.slice_obj is not None and \
                (message.slice_obj.is_client_slice() or message.slice_obj.is_broker_client_slice()):
            result = self.add_client_slice(request=message)
        else:
            super().process(message=message)
            return

        if callback_topic is None:
            self.logger.debug("No callback specified, ignoring the message")

        if self.producer.produce_sync(callback_topic, result):
            self.logger.debug("Successfully send back response: {}".format(result.to_dict()))
        else:
            self.logger.debug("Failed to send back response: {}".format(result.to_dict()))

    def get_reservations_by_category(self, *, request: GetReservationsRequestAvro,
                                     category: ReservationCategory) -> ResultReservationAvro:
        result = ResultReservationAvro()
        result.status = ResultAvro()
        try:
            if request.guid is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result = mo.get_reservations_by_category(caller=auth, category=category, id_token=request.get_id_token(),
                                                     slice_id=request.slice_id)
            result.message_id = request.message_id

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def get_slices_by_slice_type(self, *, request: GetSlicesRequestAvro, slice_type: SliceTypes) -> ResultSliceAvro:
        result = ResultSliceAvro()
        result.status = ResultAvro()
        try:
            if request.guid is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result = mo.get_slices_by_slice_type(caller=auth, slice_type=slice_type, id_token=request.get_id_token())
            result.message_id = request.message_id

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def add_client_slice(self, *, request: AddSliceAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()

        try:
            if request.guid is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result = mo.add_client_slice(caller=auth, slice_obj=request.slice_obj)
            result.message_id = request.message_id

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import unittest

from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.kernel.client_reservation_factory import ClientReservationFactory
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.reservation_holdings import ReservationHoldings


class ReservationHoldingsTest(unittest.TestCase):
    def test_create(self):
        holdings = ReservationHoldings()
        self.assertIsNotNone(holdings.list)
        self.assertIsNotNone(holdings.map)
        self.assertIsNotNone(holdings.reservation_set)

    def print_list(self, *, holdings: ReservationHoldings):
        output = ""
        for r in holdings.list:
            output += str(r)
            output += ","
        print(output)

    def make_reservation(self, *, rid: str) -> IReservation:
        return ClientReservationFactory.create(rid=ID(uid=rid))

    def check_exists(self, *, holdings: ReservationHoldings, reservation: IReservation):
        rset = holdings.get_reservations()
        return rset.contains(reservation=reservation)

    def test_add_reservation(self):
        holdings = ReservationHoldings()
        length = 5
        i = 0
        while i <= 5:
            self.assertEqual(i, holdings.size())
            res = self.make_reservation(rid=str(i))
            exist = self.check_exists(holdings=holdings, reservation=res)
            self.assertFalse(exist)
            holdings.add_reservation(reservation=res, start=5-i, end=5-i + length)
            exist = self.check_exists(holdings=holdings, reservation=res)
            self.assertTrue(exist)
            self.assertEqual(i + 1, len(holdings.list))
            self.assertEqual(i + 1, len(holdings.map))
            self.assertEqual(i + 1, holdings.reservation_set.size())
            i += 1

        res = self.make_reservation(rid="100")
        holdings.add_reservation(reservation=res, start=0, end=8)

    def test_remove_reservation(self):
        holdings = ReservationHoldings()
        length =5
        i = 0
        while i <= 5:
            self.assertEqual(i, holdings.size())
            res = self.make_reservation(rid=str(i))
            exist = self.check_exists(holdings=holdings, reservation=res)
            self.assertFalse(exist)
            holdings.add_reservation(reservation=res, start=5-i, end=5-i + length)
            exist = self.check_exists(holdings=holdings, reservation=res)
            self.assertTrue(exist)
            self.assertEqual(i + 1, len(holdings.list))
            self.assertEqual(i + 1, len(holdings.map))
            self.assertEqual(i + 1, holdings.reservation_set.size())
            i += 1
            self.print_list(holdings=holdings)

        res = self.make_reservation(rid="100")
        holdings.add_reservation(reservation=res, start=0, end=8)
        self.print_list(holdings=holdings)

        exist = self.check_exists(holdings=holdings, reservation=res)
        self.assertTrue(exist)
        self.assertEqual(i + 1, len(holdings.list))
        self.assertEqual(i + 1, len(holdings.map))
        self.assertEqual(i + 1, holdings.reservation_set.size())

        holdings.remove_reservation(reservation=res)
        exist = self.check_exists(holdings=holdings, reservation=res)
        self.assertFalse(exist)
        self.assertEqual(i, len(holdings.list))
        self.assertEqual(i, len(holdings.map))
        self.assertEqual(i, holdings.reservation_set.size())

        i = 0
        while i <= 5 :
            res = self.make_reservation(rid=str(i))
            exist = self.check_exists(holdings=holdings, reservation=res)
            self.assertTrue(exist)
            self.assertEqual(6 - i, len(holdings.list))
            self.assertEqual(6 - i, len(holdings.map))
            self.assertEqual(6 - i, holdings.reservation_set.size())

            holdings.remove_reservation(reservation=res)

            exist = self.check_exists(holdings=holdings, reservation=res)
            self.assertFalse(exist)
            self.assertEqual(5 - i, len(holdings.list))
            self.assertEqual(5 - i, len(holdings.map))
            self.assertEqual(5 - i, holdings.reservation_set.size())

            i += 1

    def test_tick(self):
        holdings = ReservationHoldings()
        length = 5
        i = 0
        while i <= 5:
            self.assertEqual(i, holdings.size())
            res = self.make_reservation(rid=str(i))
            exist = self.check_exists(holdings=holdings, reservation=res)
            self.assertFalse(exist)
            holdings.add_reservation(reservation=res, start=5-i, end=5-i + length)
            exist = self.check_exists(holdings=holdings, reservation=res)
            self.assertTrue(exist)
            self.assertEqual(i + 1, len(holdings.list))
            self.assertEqual(i + 1, len(holdings.map))
            self.assertEqual(i + 1, holdings.reservation_set.size())
            i += 1
            self.print_list(holdings=holdings)

        res = self.make_reservation(rid="100")
        holdings.add_reservation(reservation=res, start=0, end=8)
        self.print_list(holdings=holdings)

        for i in range(12):
            holdings.tick(time=i)
            size = 0
            if i < 5:
                size = 7
            else:
                size = 0

                if i == 5:
                    size = 6
                elif i == 6:
                    size = 5
                elif i == 7:
                    size = 4
                elif i == 8:
                    size = 2
                elif i == 9:
                    size = 1
            self.assertEqual(size, len(holdings.list))
            self.assertEqual(size, len(holdings.map))
            self.assertEqual(size, holdings.reservation_set.size())

    def test_intersection(self):
        holdings = ReservationHoldings()
        length = 5
        i = 0
        while i <= 5:
            self.assertEqual(i, holdings.size())
            res = self.make_reservation(rid=str(i))
            exist = self.check_exists(holdings=holdings, reservation=res)
            self.assertFalse(exist)
            holdings.add_reservation(reservation=res, start=5-i, end=5-i + length)
            exist = self.check_exists(holdings=holdings, reservation=res)
            self.assertTrue(exist)
            self.assertEqual(i + 1, len(holdings.list))
            self.assertEqual(i + 1, len(holdings.map))
            self.assertEqual(i + 1, holdings.reservation_set.size())
            i += 1
            self.print_list(holdings=holdings)

        results = [1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1, 0, 0]
        points = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]

        for i in range(len(points)):
            rset = holdings.get_reservations(time=points[i])
            self.assertIsNotNone(rset)
            self.assertEqual(results[i], rset.size())


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING, List

from fabric.actor.core.apis.i_client_reservation import IClientReservation
from fabric.actor.core.apis.i_controller_reservation import IControllerReservation
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor
from fabric.actor.core.core.actor_identity import ActorIdentity
from fabric.actor.core.core.ticket import Ticket
from fabric.actor.core.core.unit import Unit
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.time.actor_clock import ActorClock
from fabric.message_bus.messages.actor_avro import ActorAvro
from fabric.message_bus.messages.lease_reservation_avro import LeaseReservationAvro
from fabric.message_bus.messages.pool_info_avro import PoolInfoAvro
from fabric.message_bus.messages.proxy_avro import ProxyAvro
from fabric.message_bus.messages.lease_reservation_state_avro import LeaseReservationStateAvro
from fabric.message_bus.messages.reservation_mng import ReservationMng
from fabric.message_bus.messages.reservation_state_avro import ReservationStateAvro
from fabric.message_bus.messages.ticket_reservation_avro import TicketReservationAvro
from fabric.message_bus.messages.unit_avro import UnitAvro
from fabric.actor.core.proxies.actor_location import ActorLocation
from fabric.actor.core.proxies.kafka.kafka_proxy import KafkaProxy
from fabric.actor.core.proxies.local.local_proxy import LocalProxy
from fabric.actor.core.util.client import Client
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.prop_list import PropList
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.core.manage.messages.client_mng import ClientMng
from fabric.message_bus.messages.slice_avro import SliceAvro

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_reservation import IReservation
    from fabric.actor.core.apis.i_proxy import IProxy
    from fabric.actor.core.apis.i_actor import IActor


class Converter:
    @staticmethod
    def get_resource_data(*, slice_mng: SliceAvro) -> ResourceData:
        rd = ResourceData()

        rd.request_properties = slice_mng.get_request_properties()
        rd.resource_properties = slice_mng.get_resource_properties()
        rd.config_properties = slice_mng.get_config_properties()
        rd.local_properties = slice_mng.get_local_properties()
        return rd

    @staticmethod
    def absorb_res_properties(*, rsv_mng: ReservationMng, res_obj: IReservation):
        res_obj.get_resources().set_local_properties(p=
            PropList.merge_properties(incoming=rsv_mng.get_local_properties(),
                                      outgoing=res_obj.get_resources().get_local_properties()))

        res_obj.get_resources().set_config_properties(p=
            PropList.merge_properties(incoming=rsv_mng.get_config_properties(),
                                      outgoing=res_obj.get_resources().get_config_properties()))

        res_obj.get_resources().set_request_properties(p=
            PropList.merge_properties(incoming=rsv_mng.get_request_properties(),
                                      outgoing=res_obj.get_resources().get_request_properties()))

        res_obj.get_resources().set_resource_properties(p=
            PropList.merge_properties(incoming=rsv_mng.get_resource_properties(),
                                      outgoing=res_obj.get_resources().get_resource_properties()))

        return res_obj

    @staticmethod
    def fill_reservation(*, reservation: IReservation, full: bool) -> ReservationMng:
        rsv_mng = None

        if isinstance(reservation, IControllerReservation):
            rsv_mng = LeaseReservationAvro()
        elif isinstance(reservation, IClientReservation):
            rsv_mng = TicketReservationAvro()
        else:
            rsv_mng = ReservationMng()

        rsv_mng.set_reservation_id(str(reservation.get_reservation_id()))
        rsv_mng.set_slice_id(str(reservation.get_slice_id()))

        if reservation.get_type() is not None:
            rsv_mng.set_resource_type(str(reservation.get_type()))

        rsv_mng.set_units(reservation.get_units())
        rsv_mng.set_state(reservation.get_state().value)
        rsv_mng.set_pending_state(reservation.get_pending_state().value)

        if isinstance(reservation, IControllerReservation):
            rsv_mng.set_leased_units(reservation.get_leased_abstract_units())
            rsv_mng.set_join_state(reservation.get_join_state().value)
            authority = reservation.get_authority()

            if authority is not None:
                rsv_mng.set_authority(str(authority.get_guid()))

        if isinstance(reservation, IClientReservation):
            broker = reservation.get_broker()
            if broker is not None:
                rsv_mng.set_broker(str(broker.get_guid()))
            rsv_mng.set_renewable(reservation.is_renewable())
            rsv_mng.set_renew_time(reservation.get_renew_time())

        if reservation.get_term() is not None:
            rsv_mng.set_start(ActorClock.to_milliseconds(when=reservation.get_term().get_start_time()))
            rsv_mng.set_end(ActorClock.to_milliseconds(when=reservation.get_term().get_end_time()))
        else:
            if reservation.get_requested_term() is not None:
                rsv_mng.set_start(ActorClock.to_milliseconds(when=reservation.get_requested_term().get_start_time()))
                rsv_mng.set_end(ActorClock.to_milliseconds(when=reservation.get_requested_term().get_end_time()))

        if reservation.get_requested_term() is not None:
            rsv_mng.set_requested_end(ActorClock.to_milliseconds(when=reservation.get_requested_term().get_end_time()))

        rsv_mng.set_notices(reservation.get_notices())

        if full:
            rsv_mng = Converter.attach_res_properties(mng=rsv_mng, reservation=reservation)

        return rsv_mng

    @staticmethod
    def attach_res_properties(*, mng: ReservationMng, reservation: IReservation):
        resource = None
        config = None
        local = None
        request = None

        if isinstance(reservation, IControllerReservation):
            if reservation.is_active():
                resource = reservation.get_leased_resources().get_resource_properties()
                config = reservation.get_leased_resources().get_config_properties()
                request = reservation.get_leased_resources().get_request_properties()
            else:
                resource = reservation.get_resources().get_resource_properties()
                config = reservation.get_resources().get_config_properties()
                request = reservation.get_resources().get_request_properties()
            local = reservation.get_resources().get_local_properties()
        else:
            rset = reservation.get_resources()
            if rset is not None:
                resource = rset.get_resource_properties()
                config = rset.get_config_properties()
                local = rset.get_local_properties()
                request = rset.get_request_properties()

        ticket = None
        rset = reservation.get_resources()

        if rset is not None:
            cs = rset.get_resources()

            if cs is not None and isinstance(cs, Ticket):
                ticket = cs.get_properties()

        mng.set_config_properties(config)
        mng.set_request_properties(request)
        mng.set_local_properties(local)
        mng.set_resource_properties(resource)

        if isinstance(mng, TicketReservationAvro):
            mng.set_ticket_properties(ticket)

        return mng

    @staticmethod
    def fill_reservation_state(*, res: dict) -> ReservationStateAvro:
        result = None
        if 'rsv_joining' in res:
            result = LeaseReservationStateAvro()
            result.set_joining(res['rsv_joining'])
            result.set_state(res['rsv_state'])
            result.set_pending_state(res['rsv_pending'])
        else:
            result = ReservationStateAvro()
            result.set_state(res['rsv_state'])
            result.set_pending_state(res['rsv_pending'])

        return result

    @staticmethod
    def fill_reservation_states(*, res_list: list) -> List[ReservationStateAvro]:
        result = []
        for r in res_list:
            rstate = Converter.fill_reservation_state(res=r)
            result.append(rstate)

        return result

    @staticmethod
    def fill_client(*, client_mng: ClientMng) -> Client:
        result = Client()
        result.set_name(name=client_mng.get_name())
        result.set_guid(guid=ID(uid=client_mng.get_guid()))
        return result

    @staticmethod
    def fill_client_mng(*, client: dict) -> ClientMng:
        result = ClientMng()
        result.set_name(name=client['clt_name'])
        result.set_guid(guid=client['clt_guid'])
        return result

    @staticmethod
    def fill_clients(*, client_list: list) -> List[ClientMng]:
        result = []
        for c in client_list:
            mng = Converter.fill_client_mng(client=c)
            result.append(mng)

        return result

    @staticmethod
    def fill_unit_mng(*, properties: dict) -> UnitAvro:
        result = UnitAvro()
        unit = Unit.create_instance(properties)
        result.properties = unit.properties
        return result

    @staticmethod
    def fill_units(*, unit_list: list) -> List[UnitAvro]:
        result = []
        for u in unit_list:
            mng = Converter.fill_unit_mng(properties=u)
            result.append(mng)

        return result

    @staticmethod
    def fill_proxy(*, proxy: IProxy) -> ProxyAvro:
        result = ProxyAvro()
        result.set_name(proxy.get_name())
        result.set_guid(str(proxy.get_guid()))

        if isinstance(proxy, LocalProxy):
            result.set_protocol(Constants.protocol_local)
        elif isinstance(proxy, KafkaProxy):
            result.set_protocol(Constants.protocol_kafka)
            result.set_kafka_topic(proxy.get_kafka_topic())

        return result

    @staticmethod
    def fill_proxies(*, proxies: list) -> List[ProxyAvro]:
        result = []
        for p in proxies:
            proxy = Converter.fill_proxy(proxy=p)
            result.append(proxy)

        return result

    @staticmethod
    def get_agent_proxy(*, mng: ProxyAvro):
        try:
            location = ActorLocation(location=mng.get_kafka_topic())
            identity = ActorIdentity(name=mng.get_name(), guid=ID(uid=mng.get_guid()))
            from fabric.actor.core.container.container import Container
            return Container.get_proxy(protocol=mng.get_protocol(), identity=identity, location=location,
                                       type=mng.get_type())
        except Exception:
            traceback.print_exc()
            return None

    @staticmethod
    def get_resource_data_from_res(*, res_mng: ReservationMng) -> ResourceData:
        rd = ResourceData()
        rd.request_properties = res_mng.get_request_properties()
        rd.resource_properties = res_mng.get_resource_properties()
        rd.local_properties = res_mng.get_local_properties()
        rd.configuration_properties = res_mng.get_config_properties()

        return rd

    @staticmethod
    def get_resource_set(*, res_mng: ReservationMng) -> ResourceSet:
        rd = Converter.get_resource_data_from_res(res_mng=res_mng)

        return ResourceSet(units=res_mng.get_units(), rtype=ResourceType(resource_type=res_mng.get_resource_type()), rdata=rd)

    @staticmethod
    def fill_actor(*, actor: IActor) -> ActorAvro:
        result = ActorAvro()
        result.set_name(actor.get_name())
        result.set_description(actor.get_description())
        result.set_type(actor.get_type().value)
        result.set_online(True)
        result.set_id(str(actor.get_guid()))
        result.set_policy_guid(str(actor.get_plugin().get_guid()))
        return result

    @staticmethod
    def fill_actors(*, act_list: list) -> List[ActorAvro]:
        result = []
        for a in act_list:
            act_mng = Converter.fill_actor(actor=a)
            result.append(act_mng)

        return result

    @staticmethod
    def fill_actor_from_db(*, properties: dict) -> ActorAvro:
        from fabric.actor.core.core.actor import Actor
        actor = Actor.create_instance(properties=properties)
        result = ActorAvro()
        result.set_description(actor.get_description())
        result.set_name(actor.get_name())
        result.set_type(actor.get_type().value)

        from fabric.actor.core.registry.actor_registry import ActorRegistrySingleton
        aa = ActorRegistrySingleton.get().get_actor(actor_name_or_guid=actor.get_name())
        result.set_online(aa is not None)

        return result

    @staticmethod
    def fill_actors_from_db(*, act_list: list) -> List[ActorAvro]:
        result = []
        for a in act_list:
            act_mng = Converter.fill_actor_from_db(properties=a)
            result.append(act_mng)

        return result

    @staticmethod
    def fill_actors_from_db_status(*, act_list: list, status: int) -> List[ActorAvro]:
        result = []
        for a in act_list:
            act_mng = Converter.fill_actor_from_db(properties=a)

            if status == 0 or (status == 1 and act_mng.get_online()) or (status == 2 and not act_mng.get_online()):
                result.append(act_mng)

        return result

    @staticmethod
    def fill_resource_pool_descriptor(*, pool: PoolInfoAvro) -> ResourcePoolDescriptor:
        rpd = ResourcePoolDescriptor()
        properties = pool.get_properties()
        rpd.reset(properties=properties)
        return rpd



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING
from fabric.actor.core.apis.i_client_policy import IClientPolicy

if TYPE_CHECKING:
    from fabric.actor.core.kernel.resource_set import ResourceSet
    from fabric.actor.core.time.term import Term
    from fabric.actor.core.util.reservation_set import ReservationSet


class IControllerPolicy(IClientPolicy):
    """
    IControllerPolicy defines the policy interface for an actor acting in the orchestrator role.
    """

    @abstractmethod
    def get_redeeming(self, *, cycle: int) -> ReservationSet:
        """
        Returns a set of reservations that must be redeemed.
        @params cycle the current cycle
        @returns reservations to redeem
        """

    @abstractmethod
    def lease_satisfies(self, *, request_resources: ResourceSet, actual_resources: ResourceSet, requested_term: Term,
                        actual_term: Term):
        """
        Checks if the resources and term received in a lease are in compliance
        with what was initially requested. The policy can prevent the application
        of the incoming update if it disagrees with it.

        @param request_resources
                   resources requested from site authority
        @param actual_resources
                   resources received from site authority
        @param requested_term
                   term requested from site authority
        @param actual_term
                   term received from site authority
        @raises Exception in case of error
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import array as arr

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import ResourcesException


class ResourceVector:
    """
    Resource Vector
    """
    def __init__(self, *, dimensions: int):
        self.vector = arr.array('L')
        for i in range(dimensions):
            self.vector.append(0)

    def _enforce_compatibility(self, *, other):
        """
        Enforce compatibility
        @param other other
        """
        if other is None:
            raise ResourcesException("other cannot be null")

        if not isinstance(other, ResourceVector):
            raise ResourcesException(Constants.invalid_argument)

        if len(other.vector) != len(self.vector):
            raise ResourcesException("mismatched dimensions")

    def add(self, *, other):
        """
        Adds the specified vector to this instance.
        @params other : vector to add
        """
        self._enforce_compatibility(other=other)
        for i in range(len(other.vector)):
            self.vector[i] += other.vector[i]

    def subtract(self, *, other):
        """
        Subtract the specified vector to this instance.
        @params other : vector to subtract
        """
        self._enforce_compatibility(other=other)
        for i in range(len(other.vector)):
            self.vector[i] -= other.vector[i]

    def multiply(self, *, times: int):
        """
        multiply the specified vector to this instance.
        @params other : vector to multiply
        """
        for i in range(len(self.vector)):
            self.vector[i] *= times

    def subtract_me_from_and_update_me(self, *, other):
        """
        Subtracts the specified vector from this instance.
        @params other : vector to subtract
        """
        self._enforce_compatibility(other=other)
        for i in range(len(other.vector)):
            self.vector[i] = other.vector[i] - self.vector[i]

    def is_positive(self):
        for i in range(len(self.vector)):
            if self.vector[i] <= 0:
                return False
        return True

    def contains(self, *, other):
        """
        Checks if this instance contains the specified vector.
        @params other : vector to check
        @returns true or false
        """
        self._enforce_compatibility(other=other)
        for i in range(len(other.vector)):
            if self.vector[i] > other.vector[i]:
                return False
        return True

    def contains_or_equals(self, *, other):
        """
        Checks if this instance contains the specified vector.
        @params other : vector to check
        @returns true or false
        """
        self._enforce_compatibility(other=other)
        for i in range(len(other.vector)):
            if self.vector[i] >= other.vector[i]:
                return False
        return True

    def will_have_negative_dimensions(self, *, other):
        self._enforce_compatibility(other=other)
        for i in range(len(other.vector)):
            if self.vector[i] - other.vector[i] < 0:
                return True
        return False

    def will_overflow_on_subtract(self, *, to_subtract, limit):
        self._enforce_compatibility(other=to_subtract)
        self._enforce_compatibility(other=limit)
        for i in range(len(to_subtract.vector)):

            if self.vector[i] - to_subtract.vector[i] > limit.vector[i]:
                return True
        return False

    def __str__(self):
        result = "("
        for i in range(len(self.vector)):
            result += ",{}".format(self.vector[i])
        result += ")"
        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import time
import unittest
from datetime import datetime

from fabric.actor.core.apis.i_broker import IBroker
from fabric.actor.core.apis.i_broker_policy import IBrokerPolicy
from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation
from fabric.actor.core.apis.i_slice import ISlice
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.resource_pool_attribute_descriptor import ResourcePoolAttributeDescriptor, \
    ResourcePoolAttributeType
from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor
from fabric.actor.core.core.broker_policy import BrokerPolicy
from fabric.actor.core.core.ticket import Ticket
from fabric.actor.core.kernel.broker_reservation_factory import BrokerReservationFactory
from fabric.actor.core.kernel.client_reservation_factory import ClientReservationFactory
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.policy.broker_simpler_units_policy import BrokerSimplerUnitsPolicy
from fabric.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.test.client_callback_helper import ClientCallbackHelper
from fabric.actor.test.core.policy.broker_policy_test import BrokerPolicyTest


class BrokerSimplerUnitsPolicyTest(BrokerPolicyTest, unittest.TestCase):
    from fabric.actor.core.container.globals import Globals
    Globals.config_file = Constants.test_broker_configuration_file

    from fabric.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().start(force_fresh=True)
    while not GlobalsSingleton.get().start_completed:
        time.sleep(0.0001)

    def get_broker_policy(self) -> IBrokerPolicy:
        policy = BrokerSimplerUnitsPolicy()
        return policy

    def get_request_from_request(self, request: IBrokerReservation, units: int, rtype: ResourceType, start: datetime, end: datetime):
        rset = ResourceSet(units=units, rtype=rtype, rdata=ResourceData())

        term = Term(start=request.get_term().get_start_time(), end=end, new_start=start)

        result = BrokerReservationFactory.create(rid=request.get_reservation_id(), resources=rset, term=term,
                                                 slice_obj=request.get_slice())

        result.set_sequence_in(sequence=request.get_sequence_in() + 1)
        return result

    def get_request(self, units: int, rtype: ResourceType, start: datetime, end: datetime):
        slice_obj = SliceFactory.create(slice_id=ID(), name="test-slice")
        rset = ResourceSet(units=units, rtype=rtype, rdata=ResourceData())
        term = Term(start=start, end=end)
        request = BrokerReservationFactory.create(rid=ID(), resources=rset, term=term, slice_obj=slice_obj)
        return request

    def get_pool_descriptor(self, rtype: ResourceType) -> ResourcePoolDescriptor:
        rd = ResourcePoolDescriptor()
        rd.set_resource_type(rtype=rtype)
        rd.set_resource_type_label(rtype_label="Pool label: {}".format(rtype))
        ad = ResourcePoolAttributeDescriptor()
        ad.set_key(value=Constants.resource_memory)
        ad.set_label(label="Memory")
        ad.set_unit(unit="MB")
        ad.set_type(rtype=ResourcePoolAttributeType.INTEGER)
        ad.set_value(value=str(1024))
        rd.add_attribute(attribute=ad)
        return rd

    def get_source(self, units: int, rtype: ResourceType, broker: IBroker, slice_obj: ISlice):
        start = broker.get_actor_clock().cycle_start_date(cycle=self.DonateStartCycle)
        end = broker.get_actor_clock().cycle_end_date(cycle=self.DonateEndCycle)
        term = Term(start=start, end=end)

        resources = ResourceSet(units=units, rtype=rtype, rdata=ResourceData())

        properties = resources.get_resource_properties()
        rd = self.get_pool_descriptor(rtype)
        resources.set_resource_properties(p=rd.save(properties=properties, prefix=None))

        delegation = broker.get_plugin().get_ticket_factory().make_delegation(units=units, term=term, rtype=rtype,
                                                                              vector=None)
        ticket = broker.get_plugin().get_ticket_factory().make_ticket(delegation=delegation, source=None)
        cs = Ticket(ticket=ticket, plugin=broker.get_plugin())
        resources.set_resources(cset=cs)

        source = ClientReservationFactory.create(rid=ID(), resources=resources, term=term, slice_object=slice_obj)

        return source

    def check_query_response(self, response: dict, count: int):
        temp = response.get(Constants.query_response, None)
        self.assertIsNotNone(temp)
        self.assertEqual(temp, Constants.query_action_discover_pools)
        temp = response.get(Constants.pools_count, None)
        self.assertIsNotNone(temp)
        self.assertEqual(count, int(temp), count)
        result = BrokerPolicy.get_resource_pools(response)
        self.assertEqual(count, len(result))
        return result

    def test_e_donate(self):
        broker = self.get_broker()
        policy = broker.get_policy()

        slice_obj = SliceFactory.create(slice_id=ID(), name="inventory_slice")
        slice_obj.set_inventory(value=True)

        for i in range(1, 10):
            rtype = ResourceType(resource_type=str(i))
            source = self.get_source(i, rtype, broker, slice_obj)
            broker.donate_reservation(reservation=source)

            self.assertEqual(i, len(policy.inventory.get_inventory()))
            pool = policy.inventory.get(resource_type=rtype)
            self.assertIsNotNone(pool)
            self.assertEqual(i, pool.get_free())
            self.assertEqual(0, pool.get_allocated())

    def test_f_query(self):
        broker = self.get_broker()
        policy = broker.get_policy()

        request = {Constants.query_action:Constants.query_action_discover_pools}
        response = policy.query(p=request)

        print(response)

        self.check_query_response(response, 0)

        slice_obj = SliceFactory.create(slice_id=ID(), name="inventory_slice")
        slice_obj.set_inventory(value=True)

        for i in range(1, 10):
            rtype = ResourceType(resource_type=str(i))
            source = self.get_source(i, rtype, broker, slice_obj)
            broker.donate_reservation(reservation=source)

            response = policy.query(p=request)
            print(response)
            self.check_query_response(response, i)

    def test_g_advanced_request(self):
        broker = self.get_broker()
        controller = self.get_controller()
        policy = broker.get_policy()
        policy.allocation_horizon = 10

        clock = broker.get_actor_clock()

        rtype = ResourceType(resource_type="1")
        proxy = ClientCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        broker_callback = ClientCallbackHelper(name=broker.get_name(), guid=broker.get_guid())
        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=broker_callback)

        slice_obj = SliceFactory.create(slice_id=ID(), name="inventory_slice")
        slice_obj.set_inventory(value=True)

        source = self.get_source(1, rtype, broker, slice_obj)
        broker.donate_reservation(reservation=source)

        cycle = 1
        broker.external_tick(cycle=cycle)
        cycle += 1

        units = 1
        start = clock.cycle_start_date(cycle=self.DonateStartCycle)
        end = clock.cycle_end_date(cycle=self.DonateEndCycle - 1)

        request = self.get_request(units, rtype, start, end)
        broker.ticket(reservation=request, callback=proxy, caller=proxy.get_identity())

        self.assertEqual(proxy.prepared, 0)
        self.assertEqual(proxy.called, 0)

        for c in range(cycle, self.DonateEndCycle):
            broker.external_tick(cycle=c)
            while broker.get_current_cycle() != c:
                time.sleep(0.001)

            while proxy.prepared != 1:
                time.sleep(0.001)

            self.assert_ticketed(request, units, rtype, start, end)
            self.assertEqual(1, proxy.prepared)

            if proxy.called > 0:
                self.assert_ticketed(proxy.get_reservation(), units, rtype, start, end)

        broker.await_no_pending_reservations()

        self.assertEqual(1, proxy.get_called())
        self.assertTrue(request.is_closed())




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.core.common.exceptions import PluginException
from fabric.actor.core.extensions.plugin import Plugin

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_container_database import IContainerDatabase


class PluginManager:
    """
    The PluginManager is responsible for registering/unregistering
    plugin descriptors. The PluginManager registers each installed
    plugin with the backend database.
    """
    def __init__(self):
        from fabric.actor.core.container.globals import GlobalsSingleton
        self.db = None
        self.logger = GlobalsSingleton.get().get_logger()
        self.initialized = False

    def initialize(self, *, db: IContainerDatabase):
        """
        Initializes the plugin manager. Called by the management layer.
        @param db ip database
        @throws Exception in case of error
        """
        if db is None:
            raise PluginException("database cannot be null")
        if not self.initialized:
            self.db = db
            self.initialized = True

    def is_registered(self, *, plugin_id: str):
        """
        Checks if the specified plugin has been already registered.
        @param plugin_id plugin id
        @return true | false
        @throws Exception in case of error
        """
        try:
            if self.db.get_plugin(plugin_id) is not None:
                return True
        except Exception as e:
            self.logger.error(e)
        return False

    def register(self, *, plugin: Plugin):
        """
        Registers this plugin. Throws an exception if an instance of this plugin
        has already been registered. Registration involves a database access.
        @param plugin plugin instance
        @throws Exception in case of error
        """
        if self.is_registered(plugin_id=plugin.get_id()):
            raise PluginException("A plugin with this id/package id already exists")
        else:
            self.db.add_plugin(plugin=plugin)

    def unregister(self, *, plugin_id: str):
        """
        Unregisters this plugin.
        @param plugin_id plugin identifier
        @throws Exception in case of error
        """
        self.db.remove_plugin(plugin_id=plugin_id)












#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from datetime import datetime

from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation
from fabric.actor.core.apis.i_client_reservation import IClientReservation
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.time.actor_clock import ActorClock
from fabric.actor.core.time.calendar.client_calendar import ClientCalendar
from fabric.actor.core.time.calendar.source_calendar import SourceCalendar
from fabric.actor.core.util.reservation_list import ReservationList
from fabric.actor.core.util.reservation_set import ReservationSet


class BrokerCalendar(ClientCalendar):
    """
    BrokerCalendar used to organize reservation information for a broker. It builds on the functionality provided by
    ClientCalendar and extends it with the following lists:
    - closing: list of reservations organized by closing time
    - requests: list of incoming requests
    - source calendars for each source reservation
    """
    def __init__(self, *, clock: ActorClock):
        """
        Constructor
        @params clock: clock factory
        """
        super().__init__(clock=clock)
        # List of reservations grouped by closing time.
        self.closing = ReservationList()
        # Reservation requests grouped by start cycle.
        self.requests = ReservationList()
        # Source reservation calendars indexed by the source reservation identifier
        # <ReservationID, SourceCalendar>
        self.sources = {}

    def remove(self, *, reservation: IReservation):
        super().remove(reservation=reservation)
        self.remove_closing(reservation=reservation)
        if isinstance(reservation, IBrokerReservation):
            self.remove_request(reservation=reservation)

            source = reservation.get_source()
            if source is not None:
                self.remove_request(reservation=reservation, source=source)
                self.remove_outlay(source=source, client=reservation)

    def remove_scheduled_or_in_progress(self, *, reservation: IReservation):
        super().remove_scheduled_or_in_progress(reservation=reservation)
        self.remove_closing(reservation=reservation)

        if isinstance(reservation, IBrokerReservation):
            self.remove_request(reservation=reservation)

            source = reservation.get_source()
            if source is not None:
                self.remove_request(reservation=reservation, source=source)

    def get_requests(self, *, cycle: int) -> ReservationSet:
        """
        Returns all client requests for the given cycle.
        @params cycle: cycle
        @returns set of reservations representing requests starting at the specified cycle
        """
        try:
            self.lock.acquire()
            return self.requests.get_reservations(cycle=cycle)
        finally:
            self.lock.release()

    def get_all_requests(self, *, cycle: int) -> ReservationSet:
        """
        Returns all client requests up the the given cycle.
        @params cycle: cycle
        @returns set of reservations representing requests with start time no later than cycle
        """
        try:
            self.lock.acquire()
            return self.requests.get_all_reservations(cycle=cycle)
        finally:
            self.lock.release()

    def add_request(self, *, reservation: IReservation, cycle: int, source: IReservation = None):
        """
        Adds a client request.

        @params reservation: client request
        @params cycle: start cycle
        @params source: source reservation
        """
        try:
            self.lock.acquire()
            if source is None:
                self.requests.add_reservation(reservation=reservation, cycle=cycle)
            else:
                calendar = self.get_source_calendar(source=source)
                calendar.extending.add_reservation(reservation=reservation, cycle=cycle)
        finally:
            self.lock.release()

    def get_request(self, *, source: IReservation, cycle: int) -> ReservationSet:
        """
        Returns the extending requests for the given source reservation.

        @params source: source reservation
        @params cycle: cycle number

        @returns set of extending reservation requests for the given source at
                the specified cycle
        """
        try:
            self.lock.acquire()
            calendar = self.get_source_calendar(source=source)
            return calendar.extending.get_reservations(cycle=cycle)
        finally:
            self.lock.release()

    def remove_request(self, *, reservation: IReservation, source: IReservation = None):
        """
        Removes the specified reservation from the requests list.
        @params reservation:  reservation to remove
        @params source: source reservation
        """
        try:
            self.lock.acquire()
            if source is not None:
                calendar = self.get_source_calendar(source=source)
                calendar.extending.remove_reservation(reservation=reservation)
            else:
                self.requests.remove_reservation(reservation=reservation)
        finally:
            self.lock.release()

    def add_outlay(self, *, source: IReservation, client: IReservation, start: datetime, end: datetime):
        """
         Adds an outlay reservation.

        @params source: source reservation
        @params client: reservation to add
        @params start: start time
        @params end: start time
        """
        try:
            self.lock.acquire()
            calendar = self.get_source_calendar(source=source)
            calendar.outlays.add_reservation(reservation=client, start=ActorClock.to_milliseconds(when=start),
                                             end=ActorClock.to_milliseconds(when=end))
        finally:
            self.lock.release()

    def remove_outlay(self, *, source: IReservation, client: IReservation):
        """
        Removes an outlay reservation.

        @params source : source reservation
        @params client : client reservation to be removed
        """
        try:
            self.lock.acquire()
            calendar = self.get_source_calendar(source=source)
            calendar.outlays.remove_reservation(reservation=client)
        finally:
            self.lock.release()

    def add_source(self, *, source: IClientReservation):
        """
        Adds a source reservation. Creates a placeholder if necessary
        and adds the reservation to the holdings list.

        @params source:  source reservation
        """
        term = None
        try:
            self.lock.acquire()
            self.get_source_calendar(source=source)
            term = source.get_term()
        finally:
            self.lock.release()
        self.add_holdings(reservation=source, start=term.get_new_start_time(), end=term.get_end_time())

    def get_source_calendar(self, *, source: IReservation) -> SourceCalendar:
        """
        Returns the outlay calendar for the given source reservation.

        @params source : source reservation

        @returns source calendar
        """
        calendar = self.sources.get(source.get_reservation_id())
        if calendar is None:
            calendar = SourceCalendar(clock=self.clock, source=source)
            self.sources[source.get_reservation_id()] = calendar
        return calendar

    def remove_source_calendar(self, *, source: IReservation):
        """
        Removes any data structures associated with a source
        reservation.

        @params source : source reservation
        """
        try:
            self.lock.acquire()
            if source.get_reservation_id() in self.sources:
                self.sources.pop(source.get_reservation_id())
        finally:
            self.lock.release()

    def get_outlays(self, *, source: IReservation, time: datetime = None) -> ReservationSet:
        """
        Returns the client reservations satisfied from the given source
        reservation at the specified time.

        @params source : source reservation
        @params time:  time instance

        @returns set of client reservations satisfied from the given source at the specified time.
        """
        try:
            self.lock.acquire()
            calendar = self.get_source_calendar(source=source)
            if time is None:
                return calendar.outlays.get_reservations()
            else:
                return calendar.outlays.get_reservations(time=ActorClock.to_milliseconds(when=time))
        finally:
            self.lock.release()

    def get_closing(self, *, cycle: int) -> ReservationSet:
        """
        Returns all reservations that need to be closed on the specified
        cycle.

        @params cycle : cycle

        @returns a set of reservations to be closed on the specified cycle
        """
        try:
            self.lock.acquire()
            return self.closing.get_all_reservations(cycle=cycle)
        finally:
            self.lock.release()

    def add_closing(self, *, reservation: IReservation, cycle: int):
        """
        Adds a reservation to be closed on the specified cycle

        @params reservation : reservation to close
        @params cycle : cycle
        """
        try:
            self.lock.acquire()
            self.closing.add_reservation(reservation=reservation, cycle=cycle)
        finally:
            self.lock.release()

    def remove_closing(self, *, reservation: IReservation):
        """
        Removes the specified reservation from the list of closing
        reservations.

        @params reservation : reservation to remove
        """
        try:
            self.lock.acquire()
            self.closing.remove_reservation(reservation=reservation)
        finally:
            self.lock.release()

    def tick(self, *, cycle: int):
        super().tick(cycle=cycle)
        try:
            self.lock.acquire()
            self.requests.tick(cycle=cycle)
            self.closing.tick(cycle=cycle)

            for calendar in self.sources.values():
                calendar.tick(cycle=cycle)
        finally:
            self.lock.release()
